format_desc_converted_markdown = " (converted to markdown)"
format_desc_source_rest_markdown = " (source reST or markdown)"

format_help_cli = "Output format for --fetch: html, src (source .rst/.md), md (converted markdown, default)"

format_available = "Available: md (default), html, src"

query_param_desc = "Search query for ADI documentation (e.g., 'ad4000', 'axi dma')"

error_format_src_not_applicable = "Error: --format src is not applicable for URL fetch."
error_format_src_requires_index_1 = "The 'src' format fetches source files (.rst/.md) from repositories,"
error_format_src_requires_index_2 = "which requires using an index from previous search results."

